"""encodings

This module implements some basic support for conversions between
ucs2 character data (stored in an array object) and various
character encodings.

This module is a temporary hack, before python 1.6 is released
and everyone can use the built-in unicode type
"""

from array import array, ArrayType

class WStringError(ValueError):
    pass

# Find a typecode for a 2 byte unsigned type
for typecode in ['B','H','I','L',None]:
    if not typecode:
        raise ValueError('No suitable typecode')
    if array(typecode).itemsize==2:
        break

# The ucs2 encoding used by the wstring module must use instances of array.array,
# with the typecode calculated above


# Find a typecode for a 4 byte signed type
for ucs4_typecode in ['b','h','i','l',None]:
    if not ucs4_typecode:
        raise ValueError('No suitable typecode')
    if array(ucs4_typecode).itemsize==4:
        break

def utf8_to_ucs2(s):
    return array(typecode,list(utf8_to_ucs4(s)))

def utf8_to_ucs4(s,_size_map=[1]*128 + [None]*64 + [2]*32 + [3]*16 + [4]*8 + [5]*4 + [6]*2 + [None]*2,
                   _lead_mask_map=[None, 0x7f, 0x1f, 0xf, 0x7, 0x3, 0x1],
                   ord=ord):
    i = 0
    n = len(s)
    a = array(ucs4_typecode)
    while i!=n:
        lead = ord(s[i])
        size = _size_map[lead]
        if not size:
            raise WStringError('Invalid UTF-8 encoding')
        c = lead & _lead_mask_map[size]
        end = i+size
        while 1:
            i = i+1
            if i>=end:
                break
            b = ord(s[i])
            if b&0xc0 != 0x80:
                raise WStringError('Invalid UTF-8 encoding')
            c = ( c<<6 ) + ( 0x3f & b )
        a.append(c)
    return a

def ucs2_to_utf8(a,chr=chr):
    return ucs4_to_utf8(array(ucs4_typecode,list(a)))

def ucs4_to_utf8(a,chr=chr):
    o = ''
    for c in a:
        if c<0:
            raise WStringError('Invalid UCS-4 character')
        if c<=0x7f:
            o = o+chr(c)
        else:
            r = ''
            mask = 0xff80
            for ceiling in [0x7f, 0x1f, 0xf, 0x7, 0x3, 0x1]:
                if c>ceiling:
                    r = chr(0x80|(c&0x3f)) + r
                    c = c>>6
                    mask = mask>>1
                else:
                    o = o + chr((mask|(c&0x3f))&0xff) + r
                    break
    return o

def latin1_to_ucs2(s):
    return array(typecode,array('B',str(s)).tolist())

def ucs2_to_latin1(ucs2):
    o = ''
    for c in ucs2:
        if c>255:
            o = o+'?'
        else:
            o = o+chr(c)
    return o

def ucs2_to_ucs2(s):
    if type(s)==ArrayType:
        return s
    if type(s)!=type([]):
        # Ack - array uses the list accessor functions, not the sequence protocol.
        s = list(s)
    return array(typecode,s)
