import time
from array import array, ArrayType
from copy import copy

import encodings

test_plan_number=21
test_plan_approval="tdickenson"

if 67305985L==array(encodings.typecode,'\001\002\003\004')[0]:
    # big endian
    pickle_format = 2
else:
    # Little-endian
    pickle_format = 3


class wstring:
    """wstring; Immutable wide string class.
    
    Instances also look like a sequence of ucs-2 characters.

    This class is not intended to be a general purpose wide-character string.
    It's only use is to be well behaved in dtml and other bits of the zope framework:
    * Conversion to string gives a latin-1 representation. This makes wstring objects
      safe to use as the 'title' attribute, and they look right in the management
      pages.
    * Encodings are cached; this is a significant performance gain for persistant
      wide strings
    * Decodings are evaluated lazily. This is a significant performance gain for 
      wide strings that arrive in a REQUEST and go straight back to the client
      in the corresponding RESPONSE.
    * It allows <dtml-var hello_world fmt=utf8>

    When python 1.6 works well with Zope, its internal representation will change to
    the new built-in unicode type.
    """

    def __init__(self,data,encoding='guess'):
        """Initialize the wide string with the specified character data, using the given encoding.
        The internal representation is created lazily, as an optimisation for the common case where
        the only use of a wstring instance to to immediately convert it back into the encoding it
        was created from.
        """
        encoding = _encoding_aliases.get(encoding,encoding)
        if encoding=='guess':
            if isinstance(data,wstring):
                # Efficient copy
                self._init_from_sibling(data)
                return
            else:
                encoding = 'latin1'
        try:
            fn = _encoding_functions[encoding][0]
        except KeyError:
            raise encodings.WStringError('Unknown encoding',encoding)
        self._lazy_init = lambda fn=fn,data=data: fn(data)
        self._cache = { encoding: data }

    def _init_from_sibling(self,sibling):
        if hasattr(sibling,'_ucs2'):
            self._ucs2 = sibling._ucs2
        else:
            self._lazy_init = sibling._lazy_init
        self._cache = sibling._cache

    def __getattr__(self,attr):
        if '_ucs2'==attr:
            # We don't have it as an attibute? we must be lazy.
            a = self._ucs2 = self._lazy_init()
            return a
        elif known_character_encoding(attr):
            # Character encodings available as methods
            return lambda fn=self.encode, encoding=attr: fn(encoding)
        else:
            raise AttributeError(attr)

    def __hash__(self):
        return hash(self.utf8())

    def __cmp__(self,other):
        return cmp(self._ucs2, other._ucs2)

    def __nonzero__(self):
        return len(self._ucs2)>0

    def __len__(self):
        return len(self._ucs2)

    def __repr__(self):
        return "<wstring %s>" % repr(self._ucs2)[11:-1]

    def __str__(self):
        # If we get included in some dtml that is not wstring aware, then
        # we can only assume that our caller is not unicode-aware
        return self.latin1()

    def __getstate__(self):
        return pickle_format,self._ucs2.tostring()

    def __setstate__(self,state):
        old_pickle_format,s = state
        if old_pickle_format in [2,3]:
            self._ucs2 = array(encodings.typecode,s)
            if old_pickle_format != pickle_format:
                self._ucs2.byteswap()
            self._cache = {}
        else:
            raise ValueError('Unknown wide string pickle format',old_pickle_format)

    def __getitem__(self,i):
        return self._ucs2[i]

    def __getslice__(self,i,j):
        return self.__class__(self._ucs2[i:j],'ucs2')

    def encode(self,encoding):
        """Convert the string into a specified character encoding. The __getattr__ hook
        also provides method-like access to each encoding.
        """
        encoding = _encoding_aliases.get(encoding,encoding)
        e = self._cache.get(encoding,None)
        if e is None:
            try:
                fn = _encoding_functions[encoding][1]
            except KeyError:
                raise encodings.WStringError('Unknown encoding',encoding)
            e = self._cache[encoding] = fn(self._ucs2)
        return e



_encoding_aliases = { 'UTF-8':           'utf8',
                      'utf-8':           'utf8',
                      'ISO-8859-1':      'latin1',
                      'ISO_8859-1':      'latin1',
                      'ISO_8859-1:1987': 'latin1',
                    }

# A mapping from encoding id to a pair of functions.
#
# The first function converts from encoded data to ucs2, where ucs2 must
# be an instance of array.array, using the typecode from the encodings
# module. The second must be an inverse.
#
# The id needs to be usable as a ZPublisher marshalling tag, and therefore
# must be alphanumeric. More descriptive aliases can be provided in the table
# above
#
# A __getattr__ makes these ids available as method-like attributes of wstring
# instances, for example mywstring.utf8() will return the right thing.

_encoding_functions = { 'utf8':   (encodings.utf8_to_ucs2,   encodings.ucs2_to_utf8),
                        'latin1': (encodings.latin1_to_ucs2, encodings.ucs2_to_latin1),
                        'ucs2':   (encodings.ucs2_to_ucs2,   encodings.ucs2_to_ucs2),
                      }


def known_character_encoding(encoding):
    return _encoding_functions.has_key(_encoding_aliases.get(encoding,encoding))


################################################################################

class CheckError(Exception):
    pass

class InequalityError(Exception):
    pass
 
def check(t):
    if not t: raise CheckError()

def check_equality(a,*bs):
    check(len(bs))
    for b in bs:
        if a!=b: raise InequalityError(a,b)

def test():
    tester = wstring(range(0,255)+range(0xff,0xfffff,1000)+range(0xfffff,0xfffffff,100000),'ucs2')
    check_equality(wstring(tester.utf8(),'utf8'),tester)
    hangugo = wstring([0xD55C, 0xAD6D, 0xC5B4],'ucs2')
    check_equality(hangugo.utf8(),'\355\225\234\352\265\255\354\226\264')
    check_equality(wstring(hangugo.utf8(),'utf8'),hangugo)
    check_equality(hangugo.utf8(),copy(hangugo).utf8())
    check_equality(hangugo[0],0xD55C)
    check_equality(hangugo[1],0xAD6D)
    check_equality(hangugo[2],0xC5B4)
    nihongo = wstring([0x65E5, 0x672C, 0x8A9E],'ucs2')
    check_equality(wstring(nihongo.utf8(),'utf8'),nihongo)
    check_equality(nihongo.latin1(),'???')
    check_equality('hello world',wstring('hello world').latin1())
    check_equality('hello world',wstring('hello world').utf8())


if __name__=='__main__':
    print "Test-Plan-Number: %s"%test_plan_number
    print "Test-Plan-Approval: %s"%test_plan_approval
    print "Date-Performed: %s"% time.ctime(time.time())
    test()
    print "Status: PASS\n"
